module Quasar.Resources (
  -- * Resources
  Resource(..),
  dispose,

  -- * Resource management in the `Quasar` monad
  registerResource,
  registerResourceIO,
  registerDisposeAction,
  registerDisposeAction_,
  registerDisposeActionIO,
  registerDisposeActionIO_,
  registerDisposeTransaction,
  registerDisposeTransaction_,
  registerDisposeTransactionIO,
  registerDisposeTransactionIO_,
  disposeEventually,
  disposeEventually_,
  disposeEventuallyIO,
  disposeEventuallyIO_,
  captureResources,
  captureResources_,
  captureResourcesIO,
  captureResourcesIO_,

  -- * IO
  registerNewResource,
  disposeOnError,

  -- * Types to implement resources
  -- ** Disposer
  Disposer,
  TDisposer,
  disposeTDisposer,
  newUnmanagedIODisposer,
  newUnmanagedSTMDisposer,
  trivialDisposer,

  -- ** Resource manager
  ResourceManager,
  newUnmanagedResourceManagerSTM,
  attachResource,
) where


import Control.Monad.Catch
import Quasar.Future
import Quasar.Exceptions
import Quasar.MonadQuasar
import Quasar.Prelude
import Quasar.Resources.Disposer


registerResource :: (Resource a, MonadQuasar m, MonadSTM' r CanThrow m) => a -> m ()
registerResource resource = do
  rm <- askResourceManager
  liftSTM' $ attachResource rm resource
{-# SPECIALIZE registerResource :: Resource a => a -> QuasarSTM () #-}

registerResourceIO :: (Resource a, MonadQuasar m, MonadIO m) => a -> m ()
registerResourceIO res = quasarAtomically $ registerResource res
{-# SPECIALIZE registerResourceIO :: Resource a => a -> QuasarIO () #-}

registerDisposeAction :: (MonadQuasar m, MonadSTM' r CanThrow m) => IO () -> m Disposer
registerDisposeAction fn = do
  worker <- askIOWorker
  exChan <- askExceptionSink
  rm <- askResourceManager
  liftSTM' do
    disposer <- newUnmanagedIODisposer fn worker exChan
    attachResource rm disposer
    pure disposer
{-# SPECIALIZE registerDisposeAction :: IO () -> QuasarSTM Disposer #-}

registerDisposeAction_ :: (MonadQuasar m, MonadSTM' r CanThrow m) => IO () -> m ()
registerDisposeAction_ fn = liftQuasarSTM' $ void $ registerDisposeAction fn

registerDisposeActionIO :: (MonadQuasar m, MonadIO m) => IO () -> m Disposer
registerDisposeActionIO fn = quasarAtomically $ registerDisposeAction fn

registerDisposeActionIO_ :: (MonadQuasar m, MonadIO m) => IO () -> m ()
registerDisposeActionIO_ fn = quasarAtomically $ void $ registerDisposeAction fn

registerDisposeTransaction :: (MonadQuasar m, MonadSTM' r CanThrow m) => STM () -> m TDisposer
registerDisposeTransaction fn = do
  worker <- askIOWorker
  exChan <- askExceptionSink
  rm <- askResourceManager
  liftSTM' do
    disposer <- newUnmanagedSTMDisposer fn worker exChan
    attachResource rm disposer
    pure disposer
{-# SPECIALIZE registerDisposeTransaction :: STM () -> QuasarSTM TDisposer #-}

registerDisposeTransaction_ :: (MonadQuasar m, MonadSTM' r CanThrow m) => STM () -> m ()
registerDisposeTransaction_ fn = liftQuasarSTM' $ void $ registerDisposeTransaction fn

registerDisposeTransactionIO :: (MonadQuasar m, MonadIO m) => STM () -> m TDisposer
registerDisposeTransactionIO fn = quasarAtomically $ registerDisposeTransaction fn

registerDisposeTransactionIO_ :: (MonadQuasar m, MonadIO m) => STM () -> m ()
registerDisposeTransactionIO_ fn = quasarAtomically $ void $ registerDisposeTransaction fn

registerNewResource :: forall a m. (Resource a, MonadQuasar m, MonadIO m, MonadMask m) => m a -> m a
registerNewResource fn = do
  rm <- askResourceManager
  disposing <- isJust <$> peekFuture (isDisposing rm)
  -- Bail out before creating the resource _if possible_
  when disposing $ throwM AlreadyDisposing

  mask_ do
    resource <- fn
    registerResourceIO resource `catchAll` \ex -> do
      -- When the resource cannot be registered (because resource manager is now disposing), destroy it to prevent leaks
      atomically $ disposeEventually_ resource
      case ex of
        (fromException -> Just FailedToAttachResource) -> throwM AlreadyDisposing
        _ -> throwM ex
    pure resource
{-# SPECIALIZE registerNewResource :: Resource a => QuasarIO a -> QuasarIO a #-}


disposeEventuallyIO :: (Resource r, MonadIO m) => r -> m (Future ())
disposeEventuallyIO res = atomically $ disposeEventually res

disposeEventuallyIO_ :: (Resource r, MonadIO m) => r -> m ()
disposeEventuallyIO_ res = atomically $ void $ disposeEventually res


captureResources :: (MonadQuasar m, MonadSTM' r CanThrow m) => m a -> m (a, Disposer)
captureResources fn = do
  quasar <- newResourceScope
  localQuasar quasar do
    result <- fn
    pure (result, toDisposer (quasarResourceManager quasar))

captureResources_ :: (MonadQuasar m, MonadSTM' r CanThrow m) => m () -> m Disposer
captureResources_ fn = snd <$> captureResources fn

captureResourcesIO :: (MonadQuasar m, MonadIO m) => m a -> m (a, Disposer)
captureResourcesIO fn = do
  quasar <- newResourceScopeIO
  localQuasar quasar do
    result <- fn
    pure (result, toDisposer (quasarResourceManager quasar))

captureResourcesIO_ :: (MonadQuasar m, MonadIO m) => m () -> m Disposer
captureResourcesIO_ fn = snd <$> captureResourcesIO fn



-- | Runs the computation in a new resource scope, which is disposed when an exception happenes. When the computation succeeds, resources are kept.
disposeOnError :: (MonadQuasar m, MonadIO m, MonadMask m) => m a -> m a
disposeOnError fn = mask \unmask -> do
  quasar <- newResourceScopeIO
  unmask (localQuasar quasar fn) `onError` dispose quasar
